#!/usr/bin/env python3





##############################################################################
import argparse
import boto3
import datetime
import lib_error
import log
import pprint
import os
import rs_clusters
import sys
import time





##########################################################################
def generate_slice_distkey_value_lookup( error_state, rs_cluster_state, rs_connection_state ):

  if error_state.is_error():
    return

  print( 'generate_slice_distkey_value_lookup()' )

  slice_distkey_value_lookup = {}
  known_slices_count = 0

  rs_connection_state.issue_sql( error_state, sql='drop table if exists distkey_test_table;' )
  sql = '''create table distkey_test_table
           (
             column_1  int2 not null distkey
           )
           diststyle key;'''
  rs_connection_state.issue_sql( error_state, sql=sql )

  finished = False
  value = 0

  while known_slices_count < rs_cluster_state.number_slices:
    sql = 'insert into distkey_test_table ( column_1 ) values ( %d );' % (value)
    rs_connection_state.issue_sql( error_state, sql=sql )

    sql = '''select
               slice
             from
               stv_blocklist,
               pg_class
             where
               pg_class.relname     = 'distkey_test_table'  and 
               stv_blocklist.tbl    = pg_class.oid          and
               stv_blocklist.col    = 0;'''
    rows, row_count = rs_connection_state.issue_sql( error_state, sql=sql )
    slice = rows[0][0]

    if slice not in slice_distkey_value_lookup:
      slice_distkey_value_lookup[slice] = value
      known_slices_count += 1

    value += 1

    rs_connection_state.issue_sql( error_state, sql='truncate table distkey_test_table;' )

  rs_connection_state.issue_sql( error_state, sql='drop table distkey_test_table;' )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'generate_slice_distkey_value_lookup', 'Failed to create slice distkey to value lookup dictionary.' )

  return slice_distkey_value_lookup





##############################################################################
def create_and_load_base_table( error_state, rs_connection_state, s3_path, aws_access_key, aws_secret_key ):

  if error_state.is_error():
    return

  print( 'create_and_load_base_table()' )

  start = datetime.datetime.now()

  # sql = 'drop table if exists table_base;'
  # rs_connection_state.issue_sql( error_state, sql=sql )

  sql = '''create table table_base
           (
             column_1  int2        not null encode raw distkey,
             column_2  char(1024)  not null encode raw
           )
           diststyle key;'''
  rs_connection_state.issue_sql( error_state, sql=sql )

  sql = '''copy              table_base ( column_1, column_2 )
           from              '%s'
           access_key_id     '%s'
           secret_access_key '%s'
           bzip2
           compupdate        off
           csv
           maxerror          0
           statupdate        on;''' % (s3_path, aws_access_key, aws_secret_key)
  rs_connection_state.issue_sql( error_state, sql=sql )

  disk_in_use = get_total_disk_space_in_use( error_state, rs_connection_state )

  end = datetime.datetime.now()

  if error_state.is_okay():
    print( 'table_base... %s / %d' % ((end-start), disk_in_use) )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'create_and_load_base_table', 'Failed to create base table.' )

  return





##############################################################################
def create_and_populate_tables( error_state, rs_connection_state, number_tables ):

  if error_state.is_error():
    return

  print( 'create_and_populate_tables()' )

  for table_number in range ( 0, number_tables ):
    start = datetime.datetime.now()

    # sql = 'drop table if exists table_%d;' % (table_number)
    # rs_connection_state.issue_sql( error_state, sql=sql )

    sql = '''create table table_%d
             (
               column_1  int2        not null encode raw distkey,
               column_2  char(1024)  not null encode raw
             )
             diststyle key;''' % (table_number)
    rs_connection_state.issue_sql( error_state, sql=sql )

    sql = "insert into table_%d ( column_1, column_2 ) select column_1, column_2 from table_base;" % (table_number)
    rs_connection_state.issue_sql( error_state, sql=sql )

    disk_in_use = get_total_disk_space_in_use( error_state, rs_connection_state )

    end = datetime.datetime.now()

    if error_state.is_okay():
      print( 'table_%d... %s / %d' % (table_number, (end-start), disk_in_use) )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'create_and_populate_tables', 'Failed to create tables.' )

  return





##############################################################################
def drop_table( error_state, rs_connection_state, table_number ):

  if error_state.is_error():
    return

  print( 'drop_table()' )

  print( "Dropping table %d..." % (table_number) )

  sql = 'drop table table_%d;' % (table_number)
  rs_connection_state.issue_sql( error_state, sql=sql )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'drop_table', 'Failed to drop table.' )

  return





##############################################################################
def simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, text ):

  if error_state.is_error():
    return

  print( 'simple_get_most_recent_qid_for_query_text()' )

  sql = "select query from stl_querytext where sequence = 0 and text = '%s' order by query desc limit 1;" % ( text[0:200].replace("'","''") )
  rows, row_count = rs_connection_state.issue_sql( error_state, sql=sql )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'simple_get_most_recent_qid_for_query_text', 'Failed to get query ID.' )

  return rows[0][0]





##############################################################################
def get_first_scan_info_for_qid( error_state, rs_connection_state, qid ):

  if error_state.is_error():
    return

  print( 'get_first_scan_durations_and_aqua_status_for_qid()' )

  sql = "select slice, (endtime - starttime) as duration, rows, bytes, type from stl_scan where query = %d and segment = 0 and step = 0 order by slice;" % (qid)
  rows, row_count = rs_connection_state.issue_sql( error_state, sql=sql )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'get_first_scan_durations_and_aqua_status_for_qid', 'Failed to get durations and AQUA status.' )

  return rows





##########################################################################
def get_total_disk_space_in_use( error_state, rs_connection_state ):

  if error_state.is_error():
    return

  sql = 'select sum( used ) from stv_partitions where owner = host;'
  rows, row_count = rs_connection_state.issue_sql( error_state, sql=sql )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'get_total_disk_space_in_use', 'Failed to total disk space in use.' )

  return rows[0][0]





##############################################################################
def user_function_general_experiments( error_state, rs_cluster_state, rs_connection_state, results, region_code, node_type_name, node_count, user_function_args ):

  if error_state.is_error():
    return

  print( 'user_function_general_experiments()' )

  number_tables = 1

  create_and_load_base_table( error_state, rs_connection_state, user_function_args['s3_path'], user_function_args['aws_access_key'], user_function_args['aws_secret_key'] )

  create_and_populate_tables( error_state, rs_connection_state, number_tables )

  # MG2 : NO LIKE, NO SIMILAR TO

  sql = "select count(*) from table_0 where column_2 = '%s';" % ('a'*1024)
  rs_connection_state.issue_sql( error_state, sql=sql )
  qid = simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, sql )
  rows = get_first_scan_info_for_qid( error_state, rs_connection_state, qid )
  results[region_code][node_type_name]['NO LIKE, NO SIMILAR TO'] = rows

  # MG2 : LIKE

  sql = "select count(*) from table_0 where column_2 like '%a%';"
  rs_connection_state.issue_sql( error_state, sql=sql )
  qid = simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, sql )
  rows = get_first_scan_info_for_qid( error_state, rs_connection_state, qid )
  results[region_code][node_type_name]['LIKE'] = rows

  # MG2 : LIKE, COMPARE

  sql = "select count(*) from table_0 where column_2 like '%a%' and column_1 >= 1;"
  rs_connection_state.issue_sql( error_state, sql=sql )
  qid = simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, sql )
  rows = get_first_scan_info_for_qid( error_state, rs_connection_state, qid )
  results[region_code][node_type_name]['LIKE, COMPARE'] = rows

  # MG2 : LIKE, COMPARE, MAX

  sql = "select count(*), max(column_1) from table_0 where column_2 like '%a%' and column_1 >= 1;"
  rs_connection_state.issue_sql( error_state, sql=sql )
  qid = simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, sql )
  rows = get_first_scan_info_for_qid( error_state, rs_connection_state, qid )
  results[region_code][node_type_name]['LIKE, COMPARE, MAX'] = rows

  # MG2 : NO LIKE, NO SIMILAR TO, GROUP BY

  sql = "select count(*) from table_0 where column_2 = '%s' group by column_1;" % ('a'*1024)
  rs_connection_state.issue_sql( error_state, sql=sql )
  qid = simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, sql )
  rows = get_first_scan_info_for_qid( error_state, rs_connection_state, qid )
  results[region_code][node_type_name]['NO LIKE, NO SIMILAR TO, GROUP BY'] = rows

  # MG2 : LIKE, GROUP BY

  sql = "select count(*) from table_0 where column_2 like '%a%' group by column_1;"
  rs_connection_state.issue_sql( error_state, sql=sql )
  qid = simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, sql )
  rows = get_first_scan_info_for_qid( error_state, rs_connection_state, qid )
  results[region_code][node_type_name]['LIKE, GROUP BY'] = rows

  # MG2 : LIKE, DISTINCT

  sql = "select distinct column_2 from table_0 where column_2 like '%a%';"
  rs_connection_state.issue_sql( error_state, sql=sql )
  qid = simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, sql )
  rows = get_first_scan_info_for_qid( error_state, rs_connection_state, qid )
  results[region_code][node_type_name]['LIKE, DISTINCT'] = rows

  for table_number in range( 0, number_tables ):
    drop_table( error_state, rs_connection_state, table_number )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'user_function_general_experiments', 'Failed to get execute tests.' )

  return





##############################################################################
def display_results_general_experiments( error_state, results ):

  if error_state.is_error():
    return

  print( 'display_results_general_experiments()' )

  aqua_string = \
  {
    True  : '    AQUA',
    False : 'REDSHIFT'
  }

  print( '# Results (General Experiments)\n' )

  index = 0

  for region_code in results:
    print( '## %s\n' % (region_code) )
    for node_type_name in results[region_code]:
      print( '### %s\n' % (node_type_name) )
      for query_name in results[region_code][node_type_name]:
        rows = results[region_code][node_type_name][query_name]
        aqua_scan_flag = True
        index += 1
        total_rows  = 0
        total_bytes = 0
        for row in rows:
          total_rows  += row['rows']
          total_bytes += row['bytes']
          if row['type'] != 35:
            aqua_scan_flag = False
        print( '%d. %-*s : %s : %d rows : %s bytes' % (index, len('NO LIKE, NO SIMILAR TO, GROUP BY'), query_name, aqua_string[aqua_scan_flag], total_rows, total_bytes) )

  print( '\n## Raw Results (General Experiments)\n' )

  pprint.pprint( results )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'display_results_general_experiments', 'Failed to get display results.' )

  return





##############################################################################
def user_function_main_test( error_state, rs_cluster_state, rs_connection_state, results, region_code, node_type_name, node_count, user_function_args ):

  if error_state.is_error():
    return

  print( 'user_function_main_test()' )

  aqua_string = \
  {
    True  : '   AQUA',
    False : 'NO AQUA'
  }

  rms_string = \
  {
    True  : 'RMS  ',
    False : 'CACHE'
  }

  compare_type_string = \
  {
    'equality'   : 'EQUALITY  ',
    'like_%'     : 'LIKE %    ',
    'like_a'     : 'LIKE A    ',
    'like_b'     : 'LIKE B    ',
    'like_c'     : 'LIKE C    ',
    'similar to' : 'SIMILAR TO'
  }

  aqua_flag = rs_cluster_state.aqua_flag

  results[region_code][node_type_name]['aqua'] = {}
  results[region_code][node_type_name]['cluster'] = {}
  results[region_code][node_type_name]['aqua'][aqua_flag] = {}

  rc = boto3.client( 'redshift', region_code )
  results[region_code][node_type_name]['cluster'][aqua_flag] = rc.describe_clusters( ClusterIdentifier=rs_cluster_state.cluster_aws_id )

  pprint.pprint( results[region_code][node_type_name]['cluster'][aqua_flag] )

  create_and_load_base_table( error_state, rs_connection_state, user_function_args['s3_path'], user_function_args['aws_access_key'], user_function_args['aws_secret_key'] )

  for rms_flag in ( False, True ):
    results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag] = {}

    if rms_flag == True:
      number_tables = 6 # 184

    if rms_flag == False:
      number_tables = 6

    create_and_populate_tables( error_state, rs_connection_state, number_tables )

    for compare_type in ( 'equality', 'like_%', 'like_a', 'like_b', 'like_c', 'similar to' ):
      results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type] = {}

      if compare_type == 'equality':
        sql = "select count(*) from table_0 where column_2 = '%s';" % ( 'a'*1024 )

      if compare_type == 'like_%':
        sql = "select count(*) from table_1 where column_2 like '%';"

      if compare_type == 'like_a':
        sql = "select count(*) from table_2 where column_2 like '%a%';"

      if compare_type == 'like_b':
        sql = "select count(*) from table_3 where column_2 like '%b%';"

      if compare_type == 'like_c':
        sql = "select count(*) from table_4 where column_2 like '%c%';"

      if compare_type == 'similar to':
        sql = "select count(*) from table_5 where column_2 similar to '%a%';"

      for repetition in range( 0, 3 ):
        disk_used_before = get_total_disk_space_in_use( error_state, rs_connection_state )
        rs_connection_state.issue_sql( error_state, sql=sql )
        disk_used_after = get_total_disk_space_in_use( error_state, rs_connection_state )
        qid = simple_get_most_recent_qid_for_query_text( error_state, rs_connection_state, sql )
        rows = get_first_scan_info_for_qid( error_state, rs_connection_state, qid )
        results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition] = [ rows, disk_used_before, disk_used_after ]
        print( '%s : %s : %s : R%d : ( %s / %s / %s / %s )' % (aqua_string[aqua_flag], rms_string[rms_flag], compare_type_string[compare_type], repetition,
                                                               results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition][0][0][1],
                                                               results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition][0][1][1],
                                                               results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition][0][2][1],
                                                               results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition][0][3][1]) )

    for table_number in range( 0, number_tables ):
      drop_table( error_state, rs_connection_state, table_number )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'user_function_main_test', 'Failed to get execute tests.' )

  return





##############################################################################
def display_results_main_test( error_state, results ):

  if error_state.is_error():
    return

  print( 'display_results_main_test()' )

  aqua_string = \
  {
    True  : '   AQUA',
    False : 'NO AQUA'
  }

  rms_string = \
  {
    True  : 'RMS  ',
    False : 'CACHE'
  }

  compare_type_string = \
  {
    'equality'   : 'EQUALITY  ',
    'like_%'     : 'LIKE %    ',
    'like_a'     : 'LIKE A    ',
    'like_b'     : 'LIKE B    ',
    'like_c'     : 'LIKE C    ',
    'similar to' : 'SIMILAR TO'
  }

  print( '# Results (Main Test)\n' )

  for region_code in results:
    print( '## %s\n' % (region_code) )
    for node_type_name in results[region_code]:
      print( '### %s\n' % (node_type_name) )
      for aqua_flag in results[region_code][node_type_name]['aqua']:
        for rms_flag in results[region_code][node_type_name]['aqua'][aqua_flag]:
          for compare_type in results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag]:
            for repetition in results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type]:
              print( '%s : %s : %s : R%d : ( %s / %s / %s / %s )' % (aqua_string[aqua_flag], rms_string[rms_flag], compare_type_string[compare_type], repetition,
                                                                     results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition][0][0][1],
                                                                     results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition][0][1][1],
                                                                     results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition][0][2][1],
                                                                     results[region_code][node_type_name]['aqua'][aqua_flag][rms_flag][compare_type][repetition][0][3][1]) )
    print()

  print( '\n## Raw Results (Main Test)\n' )

  pprint.pprint( results )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'display_results_main_test', 'Failed to get display results.' )

  return





##############################################################################
def generate_cluster_info( error_state, s3_path=None, aws_access_key=None, aws_secret_key=None, postgres_flag=True, test_mode=True ):

  if error_state.is_error():
    return

  print( 'generate_cluster_info()' )

  node_type_names = [ 'ra3.xlplus' ]
  region_codes = [ 'us-east-1' ]

  user_function_args = \
  {
    's3_path'        : s3_path,
    'aws_access_key' : aws_access_key,
    'aws_secret_key' : aws_secret_key
  }

  # MG2 : first, spin up an AQUA cluster for some general experimental queries

  region_infos = {}

  for region_code in region_codes:
    region_infos[region_code] = {}
    if region_code in [ 'us-east-1', 'eu-west-1' ]:
      region_infos[region_code]['create_config']  = True
      region_infos[region_code]['use_defaults']   = False
      region_infos[region_code]['rs_subnet_name'] = 'rvt'
      region_infos[region_code]['rs_parameter_group_name'] = 'rvt'
      region_infos[region_code]['aqua_flag']      = True

    if region_code not in [ 'us-east-1', 'eu-west-1' ]:
      region_infos[region_code]['create_config']  = True
      region_infos[region_code]['use_defaults']   = True
      region_infos[region_code]['rs_subnet_name'] = None
      region_infos[region_code]['rs_parameter_group_name'] = 'rvt'
      region_infos[region_code]['aqua_flag']      = True

  rs_clusters_state = rs_clusters.rs_clusters( region_infos=region_infos, node_type_names=node_type_names, node_count=2, postgres_flag=postgres_flag, test_mode=test_mode )

  results = rs_clusters_state.run_clusters( error_state, user_function_general_experiments, user_function_args )

  display_results_general_experiments( error_state, results )

  # MG2 : second, the NO AQUA / AQUA test

  for aqua_flag in ( False, True ):
    region_infos = {}

    for region_code in region_codes:
      region_infos[region_code] = {}
      if region_code in [ 'us-east-1', 'eu-west-1' ]:
        region_infos[region_code]['create_config']  = True
        region_infos[region_code]['use_defaults']   = False
        region_infos[region_code]['rs_subnet_name'] = 'rvt'
        region_infos[region_code]['rs_parameter_group_name'] = 'rvt'
        region_infos[region_code]['aqua_flag']      = aqua_flag

      if region_code not in [ 'us-east-1', 'eu-west-1' ]:
        region_infos[region_code]['create_config']  = True
        region_infos[region_code]['use_defaults']   = True
        region_infos[region_code]['rs_subnet_name'] = None
        region_infos[region_code]['rs_parameter_group_name'] = 'rvt'
        region_infos[region_code]['aqua_flag']      = aqua_flag

  rs_clusters_state = rs_clusters.rs_clusters( region_infos=region_infos, node_type_names=node_type_names, node_count=2, postgres_flag=False, test_mode=test_mode )

  results = rs_clusters_state.run_clusters( error_state, user_function_main_test, user_function_args )

  display_results_main_test( error_state, results )

  error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'generate_cluster_info', 'Failed to generate Redshift cluster information.' )

  return results





##############################################################################
error_state = lib_error.error()

log.init( 'print', './fetch.log' )

ap = argparse.ArgumentParser( description='Amazon Redshift Research Project AQUA Test Script' )
sp = ap.add_subparsers( title='command', description='Tells the script what to do.', dest='command', required=True )
gi = sp.add_parser( 'run', help='Run tests.' )
gi.add_argument( '--s3_path',        type=str, required=True, help='S3 path to the test data files' )
gi.add_argument( '--aws_access_key', type=str, required=True, help='AWS access key, for COPY command' )
gi.add_argument( '--aws_secret_key', type=str, required=True, help='AWS secret key, for COPY command' )
pa = ap.parse_args()

if pa.command == 'run':
  generate_cluster_info( error_state=error_state, s3_path=pa.s3_path, aws_access_key=pa.aws_access_key, aws_secret_key=pa.aws_secret_key, postgres_flag=False, test_mode=False )

error_state.submit_on_error( lib_error.ERR_HARD, 'aqua.py', 'main', 'Failed to generate cluster information.' )

error_state.print_on_error()

log.cleanup()

print( "done" )

