##############################################################################
import lib_error
import psycopg2
import psycopg2.extras
import time





##############################################################################
class db_connection:

  ##############################################################################
  def __init__( self ):

    self.__cursor     = None
    self.__connection = None

    return

  ##############################################################################
  def connect( self, error_state, cluster_ip, cluster_port, database, username, password ):

    if error_state.is_error():
      return

    self.__cluster_ip   = cluster_ip
    self.__cluster_port = int( cluster_port )
    self.__database     = database
    self.__username     = username
    self.__password     = password

    try:
      self.__connection = psycopg2.connect( host=cluster_ip, port=cluster_port, database=database, user=username, password=password, connect_timeout=16 )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'psycopg2', 'connect', 'Connection to %s:%d failed with error "%s".' % (self.__cluster_ip, self.__cluster_port, str(e)) )

    if error_state.is_okay():
      self.__connection.autocommit = True
      self.__cursor = self.__connection.cursor( cursor_factory=psycopg2.extras.DictCursor )

      # MG2 : we can't use issue_sql(), because we'll become recursive in the event of error, as issue_sql() auto-reconnects
      #       it's a bit risky, but we issue blindly as we may be connecting to Postgres

      try:
        self.__cursor.execute( 'set enable_result_cache_for_session to off;' )
        self.__cursor.execute( 'set analyze_threshold_percent to 0;' )
        self.__cursor.execute( 'set mv_enable_aqmv_for_session to false;' )
      except:
        # error_state.submit( lib_error.ERR_HARD, 'psycopg2', 'execute', 'Failed to issue set commands.' )
        pass

    error_state.submit_on_error( lib_error.ERR_HARD, 'db_connection', 'connect', 'Failed to connect to cluster.' )

    return

  ##############################################################################
  def disconnect( self, error_state ):

    if error_state.is_error():
      return

    if self.__cursor != None:
      self.__cursor.close()
    if self.__connection != None:
      self.__connection.close()

    error_state.submit_on_error( lib_error.ERR_HARD, 'db_connection', 'disconnect', 'Failed to disconnect from cluster.' )

    return

  ##############################################################################
  def issue_sql( self, error_state, sql, print_flag=False ):

    if error_state.is_error():
      return [], -1

    if print_flag:
      print( sql )

    finished_flag          = False
    operational_error_flag = False
    retry_count            = 0
    rows                   = []
    row_count              = -1

    while error_state.is_okay() and finished_flag == False and retry_count < 5:

      try:
        self.__cursor.execute( sql )
      except psycopg2.InterfaceError as e:
        error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'InterfaceError occurred, error "%s".' % (str(e)) )
      except psycopg2.DatabaseError as e:
        error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'DatabaseError occurred, error "%s".' % (str(e)) )
      except psycopg2.DataError as e:
        error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'DataError occurred, error "%s".' % (str(e)) )
      except psycopg2.OperationalError as e:
        error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'OperationalError occurred, error "%s".' % (str(e)) )
        operational_error_flag = True
      except psycopg2.IntegrityError as e:
        error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'IntegrityError occurred, error "%s".' % (str(e)) )
      except psycopg2.InternalError as e:
        error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'InternalError occurred, error "%s".' % (str(e)) )
      except psycopg2.ProgrammingError as e:
        error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'ProgrammingError occurred, error "%s".' % (str(e)) )
      except psycopg2.NotSupportedError as e:
        error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'NotSupportedError occurred, error "%s".' % (str(e)) )

      if error_state.is_okay():
        row_count = self.__cursor.rowcount
        # MG2 : if we're a select, we may have no rows selected
        #       we could be an insert, but then we only have rows on Postgres, with "returning"
        #       this is not a good solution because it does not detect disconnection
        try:
          rows = self.__cursor.fetchall()
        except:
          rows = []
        finished_flag = True

      if error_state.is_error() and operational_error_flag == True:
        time.sleep( 1 )
        operational_error_flag = False
        retry_count += 1
        error_state.clear()
        self.connect( error_state, self.__database, self.__username, self.__password )

    if retry_count == 5:
      error_state.submit( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'Connection lost, failed to reconnect to server.' )

    error_state.submit_on_error( lib_error.ERR_HARD, 'db_connection', 'issue_sql', sql )

    error_state.submit_on_error( lib_error.ERR_HARD, 'db_connection', 'issue_sql', 'Failed to issue SQL.' )

    return rows, row_count

  ##############################################################################
  def get_column_names( self, error_state ):

    if error_state.is_error():
      return ''

    column_names = [column_info[0] for column_info in self.__cursor.description]

    error_state.submit_on_error( lib_error.ERR_HARD, 'db_connection', 'get_column_names', 'Failed to get column names.' )

    return column_names

  ##############################################################################
  def get_and_clear_notices( self, error_state ):

    if error_state.is_error():
      return ''

    # MG2 : notices have an unnecesary trailing newline
    notice = self.__connection.notices[-1][:-1]

    self.__connection.notices = []

    error_state.submit_on_error( lib_error.ERR_HARD, 'db_connection', 'get_and_clear_notices', 'Failed to get and clear notices.' )

    return notice

