##############################################################################
import boto3
import lib_error
import psycopg2
import psycopg2.extras





##############################################################################
class rs_vpc:

  ##############################################################################
  def __init__( self, region, vpc_name ):

    self.__ec2c = boto3.client( 'ec2', region )
    self.__ec2r = boto3.resource( 'ec2', region )
    self.__rc   = boto3.client( 'redshift', region )

    self.__vpc = None
    self.__ig  = None
    self.sn    = None

    self.__region            = region
    self.__availability_zone = region + ('b' if region == 'sa-east-1' else 'a')

    self.__tag_name = vpc_name

    return

  ##############################################################################
  def create( self, error_state ):

    if error_state.is_error():
      return

    # MG2 : create a VPC (this also creates a routing table, a NACL, SG and DHCP option set

    try:
      self.__vpc = self.__ec2r.create_vpc( CidrBlock         = '172.168.0.0/24',
                                           TagSpecifications = [ { 'ResourceType':'vpc', 'Tags':[{'Key':'creator', 'Value':self.__tag_name}] } ] )
      self.__vpc.wait_until_available()
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'create_vpc', 'Failed to create new VPC, %s' % (str(e)) )

    # MG2 : activate DNS support on the VPC

    try:
      self.__ec2c.modify_vpc_attribute( EnableDnsHostnames = { 'Value':True },
                                        VpcId              = self.__vpc.id )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'modify_vpc_attribute', 'Failed to modify VPC, %s' % (str(e)) )

    try:
      self.__ec2c.modify_vpc_attribute( EnableDnsSupport   = { 'Value':True },
                                        VpcId              = self.__vpc.id )
    except:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'modify_vpc_attribute', 'Failed to modify VPC, %s' % (str(e)) )

    # MG2 : create an Internet Gateweay and hook it up to the VPC

    try:
      self.__ig = self.__ec2r.create_internet_gateway( TagSpecifications = [ { 'ResourceType':'internet-gateway', 'Tags':[{'Key':'creator', 'Value':self.__tag_name}] } ] )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'create_internet_gateway', 'Failed to create Internet Gateway, %s' % (str(e)) )

    try:
      self.__vpc.attach_internet_gateway( InternetGatewayId = self.__ig.id )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'attach_internet_gateway', 'Failed to attach Internet Gateway, %s' % (str(e)) )

    # MG2 : figure out which routing table is hooked up the VPC we made, and add a route to it
    #       next() doesn't work with the iterator, so I'm using a hack

    routing_tables_iterator = self.__vpc.route_tables.all()
    for routing_table in routing_tables_iterator:
      self.__rt = routing_table
      break

    try:
      self.__route = self.__rt.create_route( DestinationCidrBlock = '0.0.0.0/0',
                                             GatewayId            = self.__ig.id,
                                             RouteTableId         = self.__rt.id )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'create_route', 'Failed to create route, %s' % (str(e)) )

    # MG2 : create a subnet and hook it up to the VPC

    try:
      self.sn = self.__ec2r.create_subnet( AvailabilityZone  = self.__availability_zone,
                                           CidrBlock         = '172.168.0.0/24',
                                           VpcId             = self.__vpc.id,
                                           TagSpecifications = [ { 'ResourceType':'subnet', 'Tags':[{'Key':'creator', 'Value':self.__tag_name}] } ] )
      self.__rsn = self.__tag_name
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'create_subnet', 'Failed to create subnet, %s' % (str(e)) )

    try:
      self.__rt.associate_with_subnet( SubnetId = self.sn.id )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'associate_with_subnet', 'Failed to associate subnet, %s' % (str(e)) )

    # MG2 : configure the subnet to auto-assign IP addresses to clusters
    #       we have to do this, because there's a default limit of seven ElasticIP addresses
    #       and it's easy to run a test which needs more than seven clusters

    try:
      self.__ec2c.modify_subnet_attribute( SubnetId            = self.sn.id,
                                           MapPublicIpOnLaunch = { 'Value':True } )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'modify_subnet_attribute', 'Failed to modify subnet, %s' % (str(e)) )

    # MG2 : find and configure the security group

    security_groups_iterator = self.__vpc.security_groups.all()
    for security_group in security_groups_iterator:
      self.__sg = security_group
      break

    try:
      self.__sg.authorize_ingress( CidrIp     = '0.0.0.0/0',
                                   IpProtocol = 'tcp',
                                   FromPort   = 5439,
                                   ToPort     = 5439 )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'authorize_ingress', 'Failed to modify security group, %s' % (str(e)) )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_vpc', 'create', 'Failed to create new VPC.' )

    return

  ##############################################################################
  def use( self, error_state ):

    if error_state.is_error():
      return

    vpc_info = ec2c.describe_vpcs(              Filters = [ {'Name':'tag:creator', 'Values':[ self.__tag_name ]} ] )
    ig_info  = ec2c.describe_internet_gateways( Filters = [ {'Name':'tag:creator', 'Values':[ self.__tag_name ]} ] )
    sn_info  = ec2c.describe_subnets(           Filters = [ {'Name':'tag:creator', 'Values':[ self.__tag_name ]} ] )

    try:
      rsn_info = self.__rc.describe_cluster_subnet_groups( ClusterSubnetGroupName = self.__tag_name )
    except:
      pass
    else:
      self.__rsn = self.__tag_name

    try:
      rpg_info = rc.describe_cluster_parameter_groups( ParameterGroupName = self.__tag_name )
    except:
      pass
    else:
      self.__rpg = self.__tag_name

    if len(vpc_info['Vpcs']) == 1:
      self.__vpc = ec2r.Vpc( vpc_info['Vpcs'][0]['VpcId'] )

    if len(ig_info['InternetGateways']) == 1:
      self.__ig = ec2r.InternetGateway( ig_info['InternetGateways'][0]['InternetGatewayId'] )

    if len(sn_info['Subnets']) == 1:
      self.sn = ec2r.Subnet( sn_info['Subnets'][0]['SubnetId'] )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_vpc', 'use', 'Failed to use existing cluster.' )

    return

  ##############################################################################
  def delete( self, error_state ):

    if error_state.is_error():
      return

    # MG2 : assuming here IG is attached...

    if self.__vpc != None and self.__ig != None:
      try:
        self.__vpc.detach_internet_gateway( InternetGatewayId = self.__ig.id )
      except Exception as e:
        error_state.submit( lib_error.ERR_HARD, 'boto3', 'detach_internet_gateway', 'Failed to detach Internet Gateway, %s' % (str(e)) )

    if self.__ig != None:
      try:
        self.__ec2c.delete_internet_gateway( InternetGatewayId = self.__ig.id )
      except Exception as e:
        error_state.submit( lib_error.ERR_HARD, 'boto3', 'delete_internet_gateway', 'Failed to delete Internet Gateway, %s' % (str(e)) )

    if self.sn != None:
      try:
        self.__ec2c.delete_subnet( SubnetId = self.sn.id )
      except Exception as e:
        error_state.submit( lib_error.ERR_HARD, 'boto3', 'delete_subnet', 'Failed to delete subnet, %s' % (str(e)) )

    if self.__rsn != None:
      try:
        self.__rc.delete_cluster_subnet_group( ClusterSubnetGroupName = self.__rsn )
      except Exception as e:
        error_state.submit( lib_error.ERR_HARD, 'boto3', 'delete_cluster_subnet_group', 'Failed to cluster subnet group, %s' % (str(e)) )

    if self.__vpc != None:
      try:
        self.__ec2c.delete_vpc( VpcId = self.__vpc.id )
      except Exception as e:
        error_state.submit( lib_error.ERR_HARD, 'boto3', 'delete_vpc', 'Failed to delete VPC, %s' % (str(e)) )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_vpc', 'delete', 'Failed to delete configuration.' )

    return

